#!$Id$

package Gtk2::FormFactory;
use Gtk2;
use Switch;
use Data::Dumper;
use strict;

my %ALIGN = (
	top	=> 0,
	middle	=> 0.5,
	bottom	=> 1,
	left	=> 0,
	center	=> 0.5,
	right	=> 1
);

=pod

=head1 NAME

Gtk2::FormFactory - a factory for creating modules

=head1 SYNOPSIS

	use Gtk2::FormFactory;

	my $factory = Gtk2::FormFactory->new;

	$factory->set_items(
		{
			type		=> 'string',
			text		=> 'Enter a text string:',
			value		=> 'Hello World!',
		},
		{
			type		=> 'boolean',
			text		=> 'Yes Or No?',
			selected	=> 1,
		},
		{
			type		=> 'file',
			text		=> 'Enter filename:',
			value		=> 'file.txt',
		},
	);

	$factory->append_item(
		{
			type		=> 'int',
			text		=> 'Enter your age:',
			value		=> 18,
			min		=> 1,
			max		=> 150,
			step		=> 10
		}
	);

	$factory->generate;

	my $widget = $factory->get_widget;

	$window->add($widget);

=head1 DESCRIPTION

Gtk2::FormFactory is designed to make it trivially easy to create forms for
managing user data, preferences, and so on. Using an array of anonymous hashes,
each of which describes an individual entry, the module creates a table
containing widgets that the user can manipulate.

=head1 CONSTRUCTOR

	my $factory = Gtk2::FormFactory->new;

This is nothing special. There are no arguments that the constructor will use.

Please note that the returned object is B<not> a Gtk2 widget. It cannot be
attached to the GUI.

=cut

sub new {
	my $self = {};
	$self->{package} = shift;
	$self->{items}    = [];
	$self->{labels}   = [];
	$self->{controls} = [];
	bless($self, $self->{package});
	return $self;
}

=pod

=head1 METHODS

	$factory->set_items(@ITEMS);

This method overrides any existing items and sets the supplied items for the
form. Each item in the C<@ITEMS> array is a reference to a hash. The hash can
contain a number of different entries:

	text	=> 'Descriptive text for the entry'

The C<text> key is used to create a label for the entry.

	type	=> 'int'

This is the data type. It can be one of:

	string
	int
	float
	enum
	file
	font
	boolean
	multiple
	text

For more information about using the different types, consult the test.pl
script. Fuller documentation may appear in the fullness of time.

=cut

sub set_items {
	my ($self, @items) = @_;
	die('usage: $factory->create_items(@items)') if (scalar(@items) < 1);
	$self->{items}    = [];
	$self->{labels}   = [];
	$self->{controls} = [];
	foreach my $item (@items) {
		$self->append_item($item);
	}
	return 1;
}

=pod

	$factory->insert_item($ITEM, $POSITION);

This inserts a new item (with the same form as those above) at position
C<$POSITION> in the list. Note that once the form has been generated, inserted
items won't appear in the form.

=cut

sub insert_item {
	my ($self, $item, $pos) = @_;
	if ($pos < 0 || $pos > scalar(@{$self->{items}})) {
		die("invalid position '$pos' argument given to insert_item");
	} else {
		my ($label, $control) = $self->create_item($item);
		splice(@{$self->{items}}, $pos, 0, $item);
		splice(@{$self->{labels}}, $pos, 0, $label);
		splice(@{$self->{controls}}, $pos, 0, $control);
		return 1;
	}
}

=pod

	$factory->append_item($ITEM);

This inserts a new item (with the same form as those above) at the bottom of
the list. Note that once the form has been generated, inserted items won't
appear in the form.

=cut

sub append_item {
	my ($self, $item) = @_;
	my ($label, $control) = $self->create_item($item);
	push(@{$self->{labels}}, $label);
	push(@{$self->{controls}}, $control);
	push(@{$self->{items}}, $item);
	return 1;
}

=pod

	$factory->prepend_item($ITEM);

This inserts a new item (with the same form as those above) at the top of the
list. Note that once the form has been generated, inserted items won't appear
in the form.

=cut

sub prepend_item {
	my ($self, $item) = @_;
	my ($label, $control) = $self->create_item($item);
	unshift(@{$self->{labels}}, $label);
	unshift(@{$self->{controls}}, $control);
	unshift(@{$self->{items}}, $item);
	return 1;
}

=pod

	$factory->set_label_align($xalign, $yalign);

Re-aligns all the labels in the form. C<$xalign> may be one of 'left', 'center'
or 'right'. C<$yalign> may be on of 'top', 'middle', or 'bottom'.

=cut

sub set_label_align {
	my ($self, $xalign, $yalign) = @_;
	foreach my $label (@{$self->{labels}}) {
		$label->set_alignment($ALIGN{$xalign}, $ALIGN{$yalign});
	}
	return 1;
}

=pod

	$factory->generate;

This processes the item data and creates a widget containing all the items.

=cut

sub generate {
	my $self = shift;
	$self->{table} = Gtk2::Table->new(2, 0, 0);
	$self->{table}->set_col_spacings(8);
	$self->{table}->set_row_spacings(8);
	$self->{table}->set_border_width(8);
	for (my $i = 0 ; $i < scalar(@{$self->{controls}}) ; $i++) {
		my $label  = @{$self->{labels}}[$i];
		my $control = @{$self->{controls}}[$i];
		$self->{table}->attach_defaults($label,   0, 1, $i, $i+1);
		$self->{table}->attach_defaults($control, 1, 2, $i, $i+1);
	}
	return 1;
}

=pod

	my $widget = $factory->get_widget;

This returns the Gtk2 widget for the form. This can be inserted into your
program's GUI.

=cut

sub get_widget {
	return $_[0]->{table};
}

=pod

	my $label = $factory->get_label($position);

Returns the label at position C<$position> in the form. This is handy if you
want to change the text, or set custom alignments.

=cut

sub get_label {
	my ($self, $pos) = @_;
	return @{$self->{labels}}[$pos];
}

=pod

	my $control = $factory->get_control($position);

This returns the control widget for the C<$position>th item. Note that this
may be a form widget, like an entry widget or spin button, or it might be a
container (in the case of a file entry).

=cut

sub get_control {
	my ($self, $pos) = @_;
	return @{$self->{controls}}[$pos];
}

=pod

	my @values = $factory->get_values;

This returns an array of all the form values as they currently exist. The order
of the values corresponds to the order of the items.

=cut

sub get_values {
	my $self = shift;
	my @values;
	for (my $i = 0 ; $i < scalar(@{$self->{items}}) ; $i++) {
		push(@values, $self->get_value($i));
	}
	return @values;
}

=pod

	my $value = $factory->get_value($position);

This returns the value of the C<$position>th form value.

=cut

sub get_value {
	my ($self, $idx) = @_;
	switch (@{$self->{items}}[$idx]->{type}) {
		case 'string' {
			return $self->get_control($idx)->get_text;
		}
		case 'int' {
			return $self->get_control($idx)->get_value;
		}
		case 'float' {
			return $self->get_control($idx)->get_value;
		}
		case 'enum' {
			my $selected_index = $self->get_control($idx)->get_history;
			return @{@{$self->{items}}[$idx]->{values}}[$selected_index];
		}
		case 'file' {
			return ($self->get_control($idx)->get_children)[0]->get_text;
		}
		case 'font' {
			return $self->get_control($idx)->child->get_text;
		}
		case 'boolean' {
			return $self->get_control($idx)->get_active;
		}
		case 'multiple' {
			my @buttons = $self->get_control($idx)->get_children;
			my @selected;
			for (my $i = 0 ; $i < scalar(@buttons) ; $i++) {
				push(@selected, $i) if ($buttons[$i]->get_active);
			}
			return @selected;
		}
		case 'text' {
			return $self->get_control($idx)->child->child->get_buffer->get_text(
				$self->get_control($idx)->child->child->get_buffer->get_start_iter,
				$self->get_control($idx)->child->child->get_buffer->get_end_iter,
				1
			);
		}
		else {
			die("unknown type");
		}
	}
}

sub create_label {
	return Gtk2::Label->new($_[1]);
}

sub create_item {
	my ($self, $item) = @_;
	my $label = $self->create_label($item->{text});
	my $control;
	switch(lc($item->{type})) {
		case 'string' {
			$control = $self->create_string_control($item->{value});
		}
		case 'int' {
			if ($item->{min} > $item->{max}) {
				die("minimum can't be less than maximum");
			} elsif ($item->{step} < 1) {
				die("step can't be less than one");
			} else {
				$control = $self->create_num_control(
					$item->{value},
					$item->{min},
					$item->{max},
					$item->{step},
					0
				);
			}
		}
		case 'float' {
			if ($item->{min} > $item->{max}) {
				die("minimum can't be less than maximum");
			} elsif ($item->{step} < 1) {
				die("step can't be less than one");
			} else {
				$control = $self->create_num_control(
					$item->{value},
					$item->{min},
					$item->{max},
					$item->{step},
					$item->{precision}
				);
			}
		}
		case 'enum' {
			if (scalar(@{$item->{values}}) < 0) {
				die("no values passed");
			} else {
				$control = $self->create_enum_control(
					$item->{values},
					$item->{value}
				);
			}
		}
		case 'file' {
			$control = $self->create_file_control($item->{value}, $item->{title}, $item->{button});
		}
		case 'font' {
			$control = $self->create_font_control($item->{value}, $item->{title});
		}
		case 'boolean' {
			$control = $self->create_boolean_control($item->{checked});
		}
		case 'multiple' {
			if (scalar(@{$item->{values}}) < 0) {
				die("no values passed");
			} else {
				$control = $self->create_multiple_control(
					$item->{values},
					$item->{selected}
				);
			}
		}
		case 'text' {
			$control = $self->create_text_control($item->{value});
		}
		else {
			die("invalid type $item->{type}");
		}
	}
	return ($label, $control);
}

sub create_string_control {
	my ($self, $str) = @_;
	my $entry = Gtk2::Entry->new;
	$entry->set_text($str);
	return $entry;
}

sub create_num_control {
	my ($self, $value, $min, $max, $step, $precision) = @_;
	my $adj = Gtk2::Adjustment->new($value, $min, $max, $step, ($step * 10), ($step * 10));
	my $control = Gtk2::SpinButton->new($adj, $step, $precision);
	return $control;
}

sub create_enum_control {
	my ($self, $values, $value) = @_;
	my $option_menu = Gtk2::OptionMenu->new;
	my $menu = Gtk2::Menu->new;
	$option_menu->set_menu($menu);
	my $i = 0;
	foreach my $text_string (@{$values}) {
		my $item = Gtk2::MenuItem->new_with_label($text_string);
		$menu->append($item);
		$option_menu->set_history($i) if ($text_string eq $value);
		$i++;
	}
	return $option_menu;
}

sub create_file_control {
	my ($self, $file, $title, $button_text) = @_;
	my $entry = Gtk2::Entry->new;
	$entry->set_text($file);
	my $button = Gtk2::Button->new($button_text);
	$button->signal_connect('clicked', sub {
		my $dialog = Gtk2::FileSelection->new($title);
		$dialog->set_filename($file);
		$dialog->signal_connect('response', sub {
			if ($_[1] == -5) {
				$entry->set_text($dialog->get_filename);
			}
			$dialog->destroy;
		});
		$dialog->run;
	});
	my $hbox = Gtk2::HBox->new;
	$hbox->set_spacing(8);
	$hbox->pack_start($entry, 1, 1, 0);
	$hbox->pack_start($button, 0, 0, 0);
	return $hbox;
}

sub create_font_control {
	my ($self, $font, $title) = @_;
	my $button = Gtk2::Button->new;
	my $label = Gtk2::Label->new;
	$button->add($label);
	my $markup = sprintf('<span font_desc="%s">%s</span>', $font, $font);
	$label->set_markup($markup);
	$button->signal_connect('clicked', sub {
		my $dialog = Gtk2::FontSelectionDialog->new($title);
		$dialog->set_font_name($font);
		$dialog->signal_connect('response', sub {
			if ($_[1] == -5) {
				$font = $dialog->get_font_name;
				$markup = sprintf('<span font_desc="%s">%s</span>', $font, $font);
				$label->set_markup($markup);
			}
			$dialog->destroy;
		});
		$dialog->run;
	});
	return $button;
}

sub create_boolean_control {
	my ($self, $checked) = @_;
	my $checkbutton = Gtk2::CheckButton->new;
	$checkbutton->set_active(1) if ($checked == 1);
	return $checkbutton;
}

sub create_multiple_control {
	my ($self, $values, $selected_values) = @_;
	my $vbox = Gtk2::VBox->new;
	$vbox->set_spacing(8);
	foreach my $text_string (@{$values}) {
		my $checkbutton = Gtk2::CheckButton->new_with_label($text_string);
		$checkbutton->set_active(1) if (grep { /^$text_string$/ } @{$selected_values} > 0);
		$vbox->pack_start($checkbutton, 0, 0, 0);
	}
	return $vbox;
}

sub create_text_control {
	my ($self, $text) = @_;
	my $textview = Gtk2::TextView->new;
	$textview->get_buffer->insert($textview->get_buffer->get_start_iter, $text);
	my $scrwin = Gtk2::ScrolledWindow->new;
	$scrwin->set_policy('automatic', 'automatic');
	$scrwin->add_with_viewport($textview);
	return $scrwin;
}

=pod

=head1 COPYRIGHT

Copyright (C) 2003, Gavin Brown. All rights reserved. This program is free
software, you can redistribute it and/or modify it under the same terms as
Perl itself. 

=head1 STATUS

This is a really early version, whether it turns into something interesting
remains to be seen.

=cut

1;
